
import subprocess
from socket import gethostbyname
from typing import List, NamedTuple, Callable, Union
from jinja2 import Environment, BaseLoader
from datetime import date
from threading import RLock
from .interfaces import ConfigurationInterface, PortDefinition
from .ssh import SSHClient
from .network.ipparser import ParsedNetworkingInformation


ValidationDefinition = NamedTuple('ValidationDefinition', [
    ('method', any), ('interval', int), ('wait_time_before_restart', int), ('kill_existing_tunnel_on_failure', bool),
    ('notify_url', str)
])


class RemotePortDefinition(PortDefinition):
    pass


class LocalPortDefinition(PortDefinition):
    pass


class Forwarding(object):
    """
    Tunnel definition local <==> remote

    Aggregate decides about SSH forwarding params
    """

    # immutable
    local: LocalPortDefinition
    remote: RemotePortDefinition
    validate: ValidationDefinition
    mode: str
    configuration: ConfigurationInterface
    retries: int
    use_autossh: bool
    health_check_connect_timeout: int
    warm_up_time: int
    time_before_restart_at_initialization: int
    wait_time_after_all_retries_failed: int

    # dynamic state
    starts_history: list
    _cache: dict

    def __init__(self, local: LocalPortDefinition,
                 remote: RemotePortDefinition,
                 validate: ValidationDefinition,
                 mode: str,
                 configuration: ConfigurationInterface,
                 retries: int,
                 use_autossh: bool,
                 health_check_connect_timeout: int,
                 warm_up_time: int,
                 time_before_restart_at_initialization: int,
                 wait_time_after_all_retries_failed: int):
        self.local = local
        self.remote = remote
        self.validate = validate
        self.mode = mode
        self.configuration = configuration
        self.retries = retries
        self.use_autossh = use_autossh
        self.health_check_connect_timeout = health_check_connect_timeout
        self.warm_up_time = warm_up_time
        self.time_before_restart_at_initialization = time_before_restart_at_initialization
        self.wait_time_after_all_retries_failed = wait_time_after_all_retries_failed

        # dynamic
        self._cache = {}
        self.starts_history = []

    def is_forwarding_remote_to_local(self):
        """
        Provide a SSH tunnel for a remote resource for connection encryption

        Case 1: (wrap remote resource into a secure tunnel)
        Case 2: Expose service from a NAT of a remote network

        :return:
        """

        return self.mode == 'local'

    def is_forwarding_local_to_remote(self):
        """
        Expose local port via tunnel on remote machine
        :return:
        """

        return self.mode == 'remote'

    def create_ssh_forwarding_signature(self) -> str:
        """
        Creates a set of SSH options for forwarding
        :return:
        """

        if 'create_ssh_forwarding' in self._cache:
            return self._cache['create_ssh_forwarding']

        c_str = ' -o ServerAliveInterval=15 -o ServerAliveCountMax=4 -o ExitOnForwardFailure=yes '

        if self.remote.gateway or self.local.gateway:
            c_str += ' -g '

        if self.is_forwarding_local_to_remote():
            c_str += '-R '

            if not self.remote.gateway:
                c_str += '%s:' % self.remote.get_host()

            c_str += '%i:%s:%i' % (
                self.remote.get_port(),
                self.local.get_host(),
                self.local.get_port()
            )

        if self.is_forwarding_remote_to_local():
            c_str += '-L '

            if not self.local.gateway:
                c_str += '%s:' % self.local.get_host()

            c_str += '%i:%s:%i' % (
                self.local.get_port(),
                self.remote.get_host(),
                self.remote.get_port()
            )

        result = self.configuration.parse(c_str)
        self._cache['create_ssh_forwarding'] = result

        return result

    def create_ssh_arguments(self, with_forwarding: bool = True) -> str:
        """
        Creates full SSH arguments, including forwarding
        :return:
        """

        return self._create_ssh_connection_string(
            with_key=True,
            with_custom_opts=True,
            append=self.create_ssh_forwarding_signature() if with_forwarding else ''
        )

    def _create_ssh_connection_string(self, with_key: bool = True, with_custom_opts: bool = True,
                                      append: str = '') -> str:
        return self.configuration.create_ssh_connection_string(
            with_key=with_key,
            with_custom_opts=with_custom_opts,
            append=append
        )

    def on_tunnel_started(self):
        self.starts_history.append(date.today())

    @property
    def current_restart_count(self):
        return len(self.starts_history) - 1 if self.starts_history else 0

    def __str__(self) -> str:
        """ Visual representation for health checks and web gui for the human """

        return 'Forwarding of [%s <-> %s] for %s' % (
            self.local, self.remote, self.configuration
        )

    @property
    def ident(self) -> str:
        return 'Forward[' + self.local.ident + '][' + self.remote.ident + ']_at_' + self.configuration.ident


class HostTunnelDefinitions(ConfigurationInterface):
    """
    Single host, multiple tunneling definitions.
    Defines ACCESS to the host, where the SSH Server is placed at.
    """

    remote_user: str
    remote_host: str
    remote_port: int
    remote_key: str
    remote_password: str
    remote_passphrase: str
    ssh_opts: str
    forward: List[Forwarding]
    variables_post_processor: Callable
    restart_all_on_forward_failure: bool
    _ip_route: Union[ParsedNetworkingInformation, None]
    _ssh: Union[SSHClient, None]
    _cache: dict
    _lock: RLock

    def __init__(self):
        self._cache = {}
        self._ssh = None
        self._lock = RLock(timeout=120)
        self._ip_route = None

    def post_process_variables(self, variables: dict) -> dict:
        if self.variables_post_processor:
            return self.variables_post_processor(variables, self)

        return variables

    def parse(self, conn_string: str):
        """
        Parses connection string ex. {{ remote_gw }}:3306 into 192.168.1.2:3306

        :param conn_string:
        :return:
        """

        to_inject = {
            'local_gw': self.get_local_gateway()
        }

        to_inject = self.post_process_variables(to_inject)

        # make it lazy
        lazy_vars = {
            'remote_gw': self.get_remote_gateway,
            'remote_interface_gw': self.get_remote_interface_gateway,
            'remote_docker_host': self.get_remote_docker_host_ip,
            'remote_docker_container': self.get_remote_interface_gateway,
            'remote_interface_eth0': lambda: self.get_remote_interface_ip('eth0'),
            'remote_interface_eth1': lambda: self.get_remote_interface_ip('eth1'),
            'remote_interface_eth2': lambda: self.get_remote_interface_ip('eth2')
        }

        for key, callback in lazy_vars.items():
            if key in conn_string and (key not in to_inject or to_inject[key] == ''):
                to_inject[key] = callback()
            else:
                to_inject[key] = ''

        tpl = Environment(loader=BaseLoader, autoescape=False).from_string(conn_string)

        return tpl.render(**to_inject)

    def get_remote_interface_ip(self, name: str):
        return self._cached(
            'get_remote_interface_ip_(%s)' % name,
            lambda: self._get_ssh_client().get_interface_ip(name)
        )

    def get_remote_interface_gateway(self):
        return self._cached(
            'get_remote_interface_gateway',
            lambda: self._get_ssh_client().get_route_gateway()
        )

    def get_remote_gateway(self):
        return self._cached(
            'get_remote_gateway',
            lambda: gethostbyname(self.remote_host).strip()
        )

    def get_remote_docker_host_ip(self):
        return self._cached(
            'get_remote_docker_host_ip',
            lambda: self._get_ssh_client().get_docker_host_ip()
        )

    def get_local_gateway(self):
        return self._cached(
            'get_local_gateway',
            lambda: self._get_parsed_ip_route().gateway_interface_ip
        )

    def _get_parsed_ip_route(self) -> ParsedNetworkingInformation:
        if self._ip_route is None:
            self._ip_route = ParsedNetworkingInformation(
                subprocess.check_output('ip route', shell=True).decode('utf-8'))

        return self._ip_route

    def _cached(self, cache_id: str, callback: Callable) -> any:
        with self._lock:
            if cache_id not in self._cache:
                self._cache[cache_id] = callback()

            return self._cache[cache_id]

    def ssh_kill_all_sessions_on_remote(self):
        with self._lock:
            self._get_ssh_client().kill_all_sessions()

    def exec_ssh(self, cmd: str, env: dict = None) -> str:
        """
        Execute a command via SSH
        :param cmd:
        :param env:
        :return:
        """

        ssh = self._get_ssh_client()

        with self._lock:
            return ssh.exec(cmd, env=env)

    def _get_ssh_client(self) -> SSHClient:
        """
        RAW ssh client, DO NOT USE - have not implemented locking
        Only for internal model usage

        :return:
        """

        with self._lock:
            if not self._ssh:
                self._ssh = SSHClient(
                    host=self.remote_host, port=self.remote_port, user=self.remote_user,
                    key=self.remote_key, password=self.remote_password, passphrase=self.remote_passphrase
                )

            return self._ssh

    def create_ssh_connection_string(self, with_key: bool = True, with_custom_opts: bool = True,
                                     append: str = '', ssh_executable: str = '') -> str:
        opts = ssh_executable

        # custom ssh options defined in the configuration file
        if self.ssh_opts and with_custom_opts:
            opts += ' ' + self.ssh_opts + ' '

        # private key to use
        if self.remote_key and with_key:
            opts += ' -i %s' % self.remote_key

        # custom string that could be passed optionally
        opts += ' ' + append + ' '

        # -p user@host, basic information
        opts += '-p %i %s@%s' % (
            self.remote_port,
            self.remote_user,
            self.remote_host
        )

        return opts

    def create_ssh_keyscan_command(self, executable: str = 'ssh-keyscan'):
        return '%s -p %i %s' % (
            executable,
            self.remote_port,
            self.remote_host
        )

    def create_complete_command_with_supervision(self, forwarding: Forwarding):
        cmd = ''
        args = forwarding.create_ssh_arguments()

        if self.remote_password:
            cmd += 'sshpass -p "%s" ' % self.remote_password

        if forwarding.use_autossh:
            cmd += "autossh -M 0 -N -f -o 'PubkeyAuthentication=yes' -nT " + args
        else:
            cmd += 'ssh -N -T ' + args

        return cmd




from requests import post as http_post
from json import dumps as json_dumps
from .model import Forwarding
from .logger import Logger


class Notify:
    @staticmethod
    def notify(fw: Forwarding, msg: str):
        if not fw.validate.notify_url:
            return

        try:
            response = http_post(
                url=fw.validate.notify_url,
                data=json_dumps({
                    'text': msg
                })
            )

            assert response.status_code == 200
        except Exception as e:
            Logger.warning('Webhook error, cannot post to "%s". Details: %s' % (fw.validate.notify_url, str(e)))

    @staticmethod
    def notify_tunnel_restarted(fw: Forwarding):
        if fw.current_restart_count == 0:
            return

        Notify.notify(fw, ':warning: The tunnel "%s" was restarted, current restart count is %i' % (
            str(fw), fw.current_restart_count
        ))




import paramiko
import subprocess

REMOTE_USER = 'proxyuser'
REMOTE_HOST = 'remote-host.org'
REMOTE_PORT = 22
REMOTE_KEY = '~/.ssh/id_rsa'
SSH_OPTS = ''


def vars_post_processor(variables: dict, host_configuration):
    # vars_post_processor() is a recognized name, just like a configuration variable
    # it is launched when local.host, remote.host and overall ssh arguments are processed
    #
    # How it works?
    # 1. You define variables there
    # 2. The templating system in TunMan executes your function and uses your variables

    # host_configuration: tunman.model.HostTunnelDefinitions
    # (not doing the typing, as you can run TunMan without installing it as a package)

    ssh: paramiko.SSHClient = host_configuration.ssh()

    # executes on REMOTE SSH
    # get IP address of a container that is named "backup_primary_db_mysql_1"
    # (this command should work with Alpine Linux and sh)
    stdin, stdout, stderr = ssh.exec_command(
        "getent hosts backup_primary_db_mysql_1 | awk '{ print $1 }'")

    variables['remote_mysql_container_ip'] = stdout.read().decode('utf-8').strip()

    return variables


# typing:
# tunnel_definition: tunman.model.Forwarding
# host_configuration: tunman.model.HostTunnelDefinitions

def mysql_check(tunnel_definition, host_configuration):
    # will raise "subprocess.CalledProcessError" on error
    subprocess.check_output('mysql -u root -proot -h %s -e "SELECT 1;"' % tunnel_definition.local.get_host(),
                            shell=True)

    return True


FORWARD = [
    {
        'local': {
            'gateway': True,      # If you want to bind to a gateway interface (to publish tunnel to the internet)
            'host': None,         # In local network an interface IP address or host you would like to bind to
            'port': 3306          # Port to bind to
        },
        'remote': {
            'gateway': False,     # Bind to a gateway interface (ssh host visible from internet)
            'host': '{{ remote_mysql_container_ip }}',  # Remote IP address is discovered by vars_post_processor()
            'port': 3306          # Port reachable on the remote host
        },
        'validate': {
            'method': mysql_check,                   # Just a check if port is open, you can place there a callback
            'interval': 60,                          # Interval for checking the health and general status
            'wait_time_before_restart': 60,          # After failure wait this time before doing restart,
                                                     # maybe the tunnel will be back
            'kill_existing_tunnel_on_failure': True  # Exit existing tunnel if it is not working
            # 'notify_url': 'http://some-slack-webhook-url'
        },
        'mode': 'local'  # Forward remote service to be visible locally (options: local, remote)
    }
]




import os
from importlib.machinery import SourceFileLoader
from typing import List
from .settings import Config
from .exceptions import ConfigurationError
from .model import HostTunnelDefinitions, Forwarding, LocalPortDefinition, RemotePortDefinition, ValidationDefinition
from .logger import Logger


class ConfigurationFactory(object):
    """
    Factory method for the model
    """

    _definitions: list

    def __init__(self, config: Config):
        self._definitions = []
        self._load_from_directory(config.CONFIG_PATH + '/conf.d/')

    def _load_from_directory(self, path: str):
        Logger.debug('Looking up configuration at "%s" path' % path)

        if not os.path.isdir(path):
            raise NotADirectoryError('Specified directory "%s" does not exist' % path)

        files = os.scandir(path)

        for file in files:
            conf_path = path + '/' + file.name

            if not os.path.isfile(conf_path):
                continue

            raw_cfg = SourceFileLoader("Conf", conf_path).load_module()

            try:
                self._definitions.append(self._parse(raw_cfg))
            except AttributeError as e:
                raise ConfigurationError('Error while parsing "%s". %s' % (conf_path, str(e)))

    def provide_all_configurations(self) -> List[HostTunnelDefinitions]:
        return self._definitions

    def _parse(self, raw) -> HostTunnelDefinitions:
        raw_opts = dir(raw)

        definition = HostTunnelDefinitions()
        definition.remote_host = raw.REMOTE_HOST
        definition.remote_port = raw.REMOTE_PORT
        definition.remote_user = raw.REMOTE_USER
        definition.remote_key = raw.REMOTE_KEY if 'REMOTE_KEY' in raw_opts else None
        definition.remote_passphrase = raw.REMOTE_KEY_PASSPHRASE if 'REMOTE_KEY_PASSPHRASE' in raw_opts else None
        definition.remote_password = raw.REMOTE_PASSWORD if 'REMOTE_PASSWORD' in raw_opts else None
        definition.forward = self._parse_forwarding(raw, configuration=definition)
        definition.variables_post_processor = raw.vars_post_processor if 'vars_post_processor' in raw_opts else None
        definition.restart_all_on_forward_failure = raw.RESTART_ALL_TUNNELS_ON_FORWARDING_FAILURE \
            if 'RESTART_ALL_TUNNELS_ON_FORWARDING_FAILURE' in raw_opts else False
        definition.ssh_opts = raw.SSH_OPTS

        return definition

    @staticmethod
    def _parse_forwarding(raw, configuration: HostTunnelDefinitions) -> List[Forwarding]:
        definitions = []

        for raw_definition in raw.FORWARD:
            definitions.append(Forwarding(
                local=LocalPortDefinition(
                    gateway=raw_definition.get('local').get('gateway', False),
                    port=raw_definition.get('local').get('port'),
                    host=raw_definition.get('local').get('host', ''),
                    configuration=configuration
                ),
                remote=RemotePortDefinition(
                    gateway=raw_definition.get('remote').get('gateway', False),
                    host=raw_definition.get('remote').get('host'),
                    port=raw_definition.get('remote').get('port'),
                    configuration=configuration
                ),
                validate=ValidationDefinition(
                    method=raw_definition.get('validate').get('method', 'none'),
                    interval=raw_definition.get('validate').get('interval', 300),
                    wait_time_before_restart=raw_definition.get('validate').get('wait_time_before_restart', 10),
                    kill_existing_tunnel_on_failure=raw_definition.get('validate').get(
                        'kill_existing_tunnel_on_failure', False),
                    notify_url=raw_definition.get('validate').get('notify_url', '')
                ),
                mode=raw_definition.get('mode'),
                configuration=configuration,
                retries=raw_definition.get('retries', 10),
                use_autossh=raw_definition.get('use_autossh', False),
                health_check_connect_timeout=raw_definition.get('health_check_connect_timeout', 60),
                warm_up_time=raw_definition.get('warm_up_time', 5),
                time_before_restart_at_initialization=raw_definition.get('time_before_restart_at_initialization', 10),
                wait_time_after_all_retries_failed=raw_definition.get('wait_time_after_all_retries_failed', 600)
            ))

        return definitions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The app module, containing the app factory function."""

import argparse
import os
import logging
from tornado.ioloop import IOLoop
from tornado.web import Application, StaticFileHandler

try:
    from .tunman.settings import Config
    from .tunman.app import TunManApplication
    from .tunman.views import ServeStatusHandler, ServeJsonStatus
    from .tunman.settings import ProdConfig, DevConfig
except ImportError:
    from tunman.settings import Config
    from tunman.app import TunManApplication
    from tunman.views import ServeStatusHandler, ServeJsonStatus
    from tunman.settings import ProdConfig, DevConfig


def start_application(config: Config, action: str):
    tunman = TunManApplication(config)

    try:
        if action == 'start':
            tunman.main()
            spawn_server(tunman, config.PORT, config.LISTEN, config.SECRET_PREFIX)
            return
        elif action == 'send-public-key':
            tunman.send_public_key()
            return
        elif action == 'add-to-known-hosts':
            tunman.add_to_known_hosts()
        else:
            print('Invalid command name, possible commands: start, send-public-key, add-to-known-hosts')
    except KeyboardInterrupt:
        print('[CTRL] + [C]')
    finally:
        tunman.on_application_close()


def spawn_server(tunman: TunManApplication, port: int, address: str = '', secret_prefix: str = ''):
    ServeStatusHandler.app = tunman

    prefix = '/'

    if secret_prefix:
        prefix += secret_prefix + "/"

    srv = Application([
        (r"" + prefix + "static/(.*)", StaticFileHandler, {'path': os.path.dirname(os.path.abspath(__file__)) + '/tunman/static'}),
        (r"" + prefix + "health", ServeJsonStatus),
        (r"" + prefix, ServeStatusHandler)
    ])

    # disable logger
    hn = logging.NullHandler()
    hn.setLevel(logging.DEBUG)

    if tunman.settings.DEBUG is False:
        for logger_name in ['tornado.application', 'tornado.general']:
            logging.getLogger(logger_name).addHandler(hn)
            logging.getLogger(logger_name).propagate = False

    srv.listen(port, address)
    IOLoop.current().start()


def main():
    #
    # Arguments parsing
    #
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c',
        '--config',
        help='Path to the configuration',
        default=os.getenv('TUNMAN_CONFIG', '.')
    )
    parser.add_argument(
        '-p',
        '--port',
        help='HTTP port to listen on',
        default=8015
    )
    parser.add_argument(
        '-l',
        '--listen',
        help='Address to listen on, defaults to 0.0.0.0',
        default=None
    )
    parser.add_argument(
        '-s',
        '--secret-prefix',
        default=os.getenv('TUNMAN_SECRET_PREFIX', ''),
        help='Add a subdirectory prefix to the URL example: https://your-domain.org/some-secret-code-here/health'
    )
    parser.add_argument(
        'action',
        metavar='N',
        type=str,
        help='Action. Choice: start, send-public-key, add-to-known-hosts'
    )
    parser.add_argument(
        '-e',
        '--env',
        help='Environment: debug, prod',
        default=os.getenv('TUNMAN_ENV', 'prod')
    )

    parsed = parser.parse_args()
    config = ProdConfig() if parsed.env == 'prod' else DevConfig()
    config.CONFIG_PATH = parsed.config
    config.PORT = parsed.port
    config.LISTEN = parsed.listen
    config.SECRET_PREFIX = parsed.secret_prefix

    start_application(config, parsed.action)


if __name__ == '__main__':
    main()



#
# CONFIGURATION REFERENCE
# -----------------------
#
#  This file should always contain all possible configuration options for documentation
#  It does not serve to run. The configuration should show possible options, not a well configured setup to run.
#


# ======================================================
#  Basic SSH connection details, common for all tunnels
# ======================================================
REMOTE_USER = 'proxyuser'
REMOTE_HOST = 'remote-host.org'
REMOTE_PORT = 22
REMOTE_KEY = '~/.ssh/id_rsa'
SSH_OPTS = ''

# ==========================================================================
#  Defined SSH tunnels that will be forwarded via SSH host specified above
# ==========================================================================
FORWARD = [
    {
        'local': {
            'gateway': True,      # If you want to bind to a gateway interface (to publish tunnel to the internet)
            'host': None,         # In local network an interface IP address or host you would like to bind to
            'port': 8010          # Port to bind to
        },
        'remote': {
            'gateway': False,     # Bind to a gateway interface (ssh host visible from internet)
            'host': '127.0.0.1',  # IP address of a service reachable on the remote host
            'port': 80            # Port reachable on the remote host
        },
        'validate': {
            'method': 'local_port_ping',              # Opts: local_port_ping, remote_port_ping,
                                                      #        you can place there a callback
            'interval': 60,                           # Checks tunnel health and status each X seconds
            'wait_time_before_restart': 60,           # After failure wait this time before doing restart,
                                                      # maybe the tunnel will be back without doing anything
            'kill_existing_tunnel_on_failure': True,  # Exit existing tunnel if it is not working,
            'notify_url': 'http://some-slack-webhook-url'  # Slack/Mattermost integration
        },
        'mode': 'local',       # local - forward remote resource to localhost, remote - reverse, to remote
        'retries': 15,                              # number of retries
        'wait_time_after_all_retries_failed': 600,  # time to wait, when all retries exhausted

        'use_autossh': False,  # use autossh? (not recommended), may be deprecated and removed in future releases

        'health_check_connect_timeout': 60,   # timeout for the health check
        'warm_up_time': 5,                    # wait this time before saying that the tunnel was started successfully

        'time_before_restart_at_initialization': 10,  # wait this time before restarting, when the process
                                                      # does not start from the beginning
    }
]




import psutil
import subprocess
from typing import Union, List, Tuple
from ..logger import Logger


class SystemProcessManager:
    """
    Manages all opened processes
    """

    _procs: List[subprocess.Popen]

    def __init__(self):
        self._procs = []

    """
    System process helper methods
    """

    def spawn(self, cmd: str) -> subprocess.Popen:
        Logger.info('Spawning %s' % cmd)
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.wait(proc)

        if proc.poll() is None:
            self._procs.append(proc)

        return proc

    @staticmethod
    def communicate(proc: subprocess.Popen) -> Tuple[str, str]:
        try:
            streams = proc.communicate(timeout=2)
            return streams[0].decode('utf-8'), streams[1].decode('utf-8')
        except subprocess.TimeoutExpired:
            return '', ''

    def close_all_tunnels(self, signatures: List[str]):
        """
        Kill all processes spawned previously (finds by tunnel parameters)
        :return:
        """

        for signature in signatures:
            proc = self.find_process_by_signature(signature)

            if proc:
                Logger.info('Killing %i (%s)' % (proc.pid, proc.name()))
                self._kill_proc(proc)

            for proc in psutil.process_iter():
                cmdline = " ".join(proc.cmdline())

                if signature in cmdline:
                    self._kill_proc(proc)

        for proc in self._procs:
            Logger.info('Killing %i' % proc.pid)

            self._kill_proc(proc)

    @staticmethod
    def _kill_proc(proc):
        SystemProcessManager.wait(proc)
        proc.kill()

    @staticmethod
    def wait(proc) -> bool:
        try:
            proc.wait(timeout=1)
        except subprocess.TimeoutExpired:
            return True
        except psutil.TimeoutExpired:
            return True

        return False

    @staticmethod
    def find_process_by_signature(signature: str) -> Union[psutil.Process, None]:
        for proc in psutil.process_iter():
            cmdline = " ".join(proc.cmdline())

            if signature in cmdline and "ssh" in cmdline:
                return proc

        return None

    @staticmethod
    def kill_process_by_signature(signature: str):
        proc = SystemProcessManager.find_process_by_signature(signature)

        if proc:
            proc.kill()

    def clean_up_already_exited_processes(self):
        """ Free up information about processes that no longer are alive,
            so the application will not attempt to kill when gracefully shutting down
        """

        for proc in self._procs.copy():
            Logger.debug('clean_up: Checking if process pid=%i is still alive' % proc.pid)

            if proc.poll() is not None:
                Logger.debug('clean_up: Freeing proc pid=%i' % proc.pid)

                try:
                    self._procs.remove(proc)
                except ValueError:
                    continue

    def get_procs_count(self) -> int:
        return len(self._procs)




import threading
import os
from .manager.ssh import TunnelManager
from .model import HostTunnelDefinitions
from .factory import ConfigurationFactory
from .settings import Config
from .logger import setup_logger, Logger
from time import sleep

"""
    Application main() - spawns threads managed by TunnelManager()
"""


class TunManApplication(object):
    _threads: list
    config: ConfigurationFactory
    settings: Config
    tun_manager: TunnelManager

    def __init__(self, config: Config):
        setup_logger(config.LOG_PATH, config.LOG_LEVEL)
        self.config = ConfigurationFactory(config)
        self.settings = config
        self.tun_manager = TunnelManager()
        self._threads = []

    def main(self):
        """ Start tunnelling and the web server """

        for config in self.config.provide_all_configurations():
            self._spawn_threads(config)

    def send_public_key(self):
        """ Execute ssh-copy-id for all configured hosts """

        for config in self.config.provide_all_configurations():
            Logger.info('Processing %s, please enter credentials when asked' % str(config))
            os.system(config.create_ssh_connection_string(ssh_executable='ssh-copy-id'))

    def add_to_known_hosts(self):
        """ Executes ssh-keyscan and adds signatures to ~/.ssh/known_hosts """

        os.system('mkdir -p ~/.ssh; touch ~/.ssh/known_hosts')
        path = os.path.expanduser('~/.ssh/known_hosts')

        with open(path, 'rb') as f:
            content = f.read().decode('utf-8')

        for config in self.config.provide_all_configurations():
            Logger.info('Adding %s to the %s' % (str(config), path))

            if config.remote_host in content:
                Logger.info('%s already present in the %s' % (config.remote_host, path))
                continue

            os.system(config.create_ssh_keyscan_command('ssh-keyscan') + ' >> ~/.ssh/known_hosts')

    def _spawn_threads(self, configuration: HostTunnelDefinitions):
        Logger.info('Spawning thread for %s' % configuration)

        for definition in configuration.forward:
            thr = threading.Thread(target=lambda: self.tun_manager.spawn_tunnel(definition, configuration))
            thr.start()
            self._threads.append(thr)
            sleep(0.5)

    def on_application_close(self):
        Logger.debug('Closing the application')
        self.tun_manager.close_all_tunnels()




import os
import sys
import unittest
from unittest_data_provider import data_provider

sys.path.append(os.path.dirname(__file__) + "/../tunman")

from ..tunman.network.ipparser import ParsedNetworkingInformation


def data():
    return [
        # Check local gateway interface IP address
        [
            '''
                default via 192.168.0.1 dev wlp2s0 proto dhcp metric 600 
                172.17.0.0/16 dev docker0 proto kernel scope link src 172.17.0.1 
                172.29.0.0/16 dev br-58e9d7c4f56c proto kernel scope link src 172.29.0.1 linkdown 
                192.168.0.0/24 dev wlp2s0 proto kernel scope link src 192.168.0.109 metric 600 
            ''',
            {
                'gw_interface': 'wlp2s0',
                'gw': '192.168.0.1',
                'gw_interface_ip': '192.168.0.109'
            }
        ],

        # With a VPN
        [
            '''
                default via 192.168.0.1 dev wlp2s0 proto dhcp metric 600 
                10.5.0.0/16 dev br-60c84770cf1c proto kernel scope link src 10.5.0.1 linkdown 
                10.10.74.0/24 dev tun0 proto static scope link metric 50 
                10.40.0.22 dev tun0 proto kernel scope link src 10.40.0.22 metric 50 
                10.74.73.0/24 dev tun0 proto static scope link metric 50 
                10.74.74.0/24 dev tun0 proto static scope link metric 50 
                10.74.75.0/24 dev tun0 proto static scope link metric 50 
                85.1.2.3 via 192.168.0.1 dev wlp2s0 proto static metric 600 
                172.17.0.0/16 dev docker0 proto kernel scope link src 172.17.0.1 
                172.18.0.0/16 dev br-0a61433a1ec6 proto kernel scope link src 172.18.0.1 linkdown 
                172.19.0.0/16 dev br-6392b79fc4d8 proto kernel scope link src 172.19.0.1 linkdown 
                172.21.0.0/16 dev br-00f6e41d973a proto kernel scope link src 172.21.0.1 linkdown 
                172.27.0.0/16 dev br-d3ecf485c906 proto kernel scope link src 172.27.0.1 linkdown 
                172.29.0.0/16 dev br-58e9d7c4f56c proto kernel scope link src 172.29.0.1 linkdown 
                172.31.0.0/16 dev tun0 proto static scope link metric 50 
                192.168.0.0/24 dev tun0 proto static scope link metric 50 
                192.168.0.0/24 dev wlp2s0 proto kernel scope link src 192.168.0.109 metric 600 
                192.168.0.1 dev wlp2s0 proto static scope link metric 600 
                192.168.2.0/24 dev tun0 proto static scope link metric 50 
                192.168.32.0/20 dev br-13da41dbf0cd proto kernel scope link src 192.168.32.1 linkdown 
                192.168.74.0/24 dev tun0 proto static scope link metric 50 
                192.168.96.0/20 dev br-ab66953755d3 proto kernel scope link src 192.168.96.1 linkdown 
                192.168.112.0/20 dev br-7ae3e41a3978 proto kernel scope link src 192.168.112.1 linkdown 
                192.168.160.0/20 dev br-fd5ffd8f13e2 proto kernel scope link src 192.168.160.1 linkdown
            ''',
            {
                'gw_interface': 'wlp2s0',
                'gw': '192.168.0.1',
                'gw_interface_ip': '192.168.0.109'
            }
        ]
    ]


class ParsedNetworkingInformationTest(unittest.TestCase):
    @data_provider(data)
    def test_resolves_values_correctly(self, output: str, checks: dict):
        parsed = ParsedNetworkingInformation(output)

        self.assertEqual(parsed.gateway, checks['gw'])
        self.assertEqual(parsed.gateway_interface_ip, checks['gw_interface_ip'])
        self.assertEqual(parsed.gateway_interface, checks['gw_interface'])




import psutil
import socket
from typing import Callable
from .model import Forwarding, HostTunnelDefinitions
from .logger import Logger


# @todo: Add connection timeouts
class Validation:
    @staticmethod
    def check_tunnel_alive(definition: Forwarding, configuration: HostTunnelDefinitions) -> bool:
        validation = definition.validate.method

        try:
            if type(validation) == Callable:
                return validation(definition, configuration)

            if validation == 'local_port_ping':
                return Validation.check_port_responding(
                    definition.local.get_host_as_ip_address(),
                    definition.local.port
                )
            elif validation == 'remote_port_ping':
                return Validation.check_remote_port_responding(
                    definition.remote.get_host_as_ip_address(),
                    definition.remote.port,
                    configuration
                )

        except Exception as e:
            Logger.error('Validation error:' + str(e))
            return False

        # no defined health check
        return True

    @staticmethod
    def is_process_alive(signature: str) -> bool:
        for proc in psutil.process_iter():
            cmdline = " ".join(proc.cmdline())

            if signature in cmdline:
                return proc.pid > 0

        return False

    @staticmethod
    def check_port_responding(host: str, port: int) -> bool:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(15)

        try:
            return sock.connect_ex((host, port)) == 0
        finally:
            sock.close()

    @staticmethod
    def check_remote_port_responding(host: str, port: int, configuration: HostTunnelDefinitions) -> bool:
        exit_code = int(configuration.exec_ssh('nc -zw15 %s %i 1>&2; echo $?' % (host, port)).strip())

        return exit_code == 0




import os
import json
from typing import Optional, Awaitable
from tornado.web import RequestHandler
from jinja2 import Environment, FileSystemLoader
from typing import List
from .app import TunManApplication
from .model import Forwarding


class ServeStatusHandler(RequestHandler):
    app: TunManApplication = None

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def get(self):
        loader = FileSystemLoader(os.path.dirname(os.path.abspath(__file__)) + '/templates')
        tpl = Environment(loader=loader, autoescape=False).get_template('status.html.j2')

        self.write(tpl.render(**self._get_data()))

    def _get_data(self) -> dict:
        all_forwardings = self._get_forwardings()
        stats = self.app.tun_manager.get_stats(all_forwardings)
        data = {
            'forwardings': []
        }

        for definition in all_forwardings:
            forwarding = {
                'is_alive': False,
                'current_pid': '',
                'ident': definition.ident,
                'signature': definition.create_ssh_forwarding_signature(),
                'restarts_count': 0
            }

            if definition in stats['status']:
                forwarding['is_alive'] = stats['status'][definition]['is_alive']
                forwarding['current_pid'] = stats['status'][definition]['pid']
                forwarding['restarts_count'] = stats['status'][definition]['restarts_count']

            data['forwardings'].append(forwarding)

        return data

    def _get_forwardings(self) -> List[Forwarding]:
        all_definitions = []

        for host_tunnels in self.app.config.provide_all_configurations():
            for tunnel in host_tunnels.forward:
                all_definitions.append(tunnel)

        return all_definitions


class ServeJsonStatus(ServeStatusHandler):
    def get(self):
        """ Returns a JSON formatted status page """

        data = self._get_data()
        tunnels = {}
        global_status = True

        for forwarding in data['forwardings']:
            if not forwarding['is_alive']:
                global_status = False

            tunnels[forwarding['ident']] = {
                'ok': forwarding['is_alive'],
                'ident': forwarding['ident'] + '=' + str(forwarding['is_alive'])
            }

        self.add_header('Content-Type', 'application/json')
        self.write(
            json.dumps({
                'status': {
                    'tunnels': tunnels,
                    'ident': 'global_status=' + str(global_status),
                    'ok': global_status
                },
                'data': data
            }, indent=4)
        )



# -*- coding: utf-8 -*-
"""Application configuration."""
import os


class Config(object):
    """Base configuration."""

    APP_DIR = os.path.abspath(os.path.dirname(__file__))  # This directory
    PROJECT_ROOT = os.path.abspath(os.path.join(APP_DIR, os.pardir))
    CONFIG_PATH = os.getenv('CONFIG_PATH', os.path.abspath(os.path.dirname(__file__)))
    LOG_LEVEL = 'info'
    LOG_PATH = './tunman.log'
    SECRET_PREFIX = ''


class ProdConfig(Config):
    """Production configuration."""

    ENV = 'prod'
    DEBUG = False


class DevConfig(Config):
    """Development configuration."""

    ENV = 'dev'
    DEBUG = True
    LOG_LEVEL = 'debug'




import abc
from typing import Callable


class ConfigurationInterface(abc.ABC):
    remote_user: str
    remote_host: str
    remote_port: int
    remote_key: str
    remote_password: str
    remote_passphrase: str
    forward: list
    ssh_opts: str
    variables_post_processor: Callable
    restart_all_on_forward_failure: bool

    @abc.abstractmethod
    def post_process_variables(self, variables: dict) -> dict:
        pass

    @abc.abstractmethod
    def parse(self, conn_string: str) -> str:
        pass

    @abc.abstractmethod
    def get_remote_gateway(self) -> str:
        pass

    @abc.abstractmethod
    def get_remote_docker_host_ip(self) -> str:
        pass

    @abc.abstractmethod
    def get_local_gateway(self):
        pass

    @abc.abstractmethod
    def create_ssh_connection_string(self, with_key: bool = True, with_custom_opts: bool = True,
                                     append: str = '') -> str:
        pass

    @abc.abstractmethod
    def create_ssh_keyscan_command(self, executable: str = 'ssh-keyscan') -> str:
        pass

    def __str__(self):
        return 'Host<ssh=%s@%s:%i> (contains %i forwardings)' % (
            self.remote_user,
            self.remote_host,
            self.remote_port,
            len(self.forward)
        )

    @property
    def ident(self) -> str:
        return self.remote_user + '@' + self.remote_host + ':' + str(self.remote_port)


class PortDefinition(object):
    gateway: bool
    host: str
    port: int
    configuration: ConfigurationInterface

    def __init__(self, gateway: bool, host: str, port: int, configuration: ConfigurationInterface):
        self.gateway = gateway
        self.host = host
        self.port = port
        self.configuration = configuration

    def __str__(self):
        return 'PortDefinition<%s:%i, gw=%s>' % (
            self.host, self.port, str(self.gateway)
        )

    def get_host(self):
        host = self.host

        if not host:
            return '0.0.0.0'

        return self.configuration.parse(host)

    def get_host_as_ip_address(self):
        host = self.get_host()

        if host == '*':
            host = '0.0.0.0'

        return host

    def get_port(self) -> int:
        return int(self.configuration.parse(str(self.port)))

    @property
    def ident(self):
        ident = self.host + ':' + str(self.port)

        if self.gateway:
            ident += '@gw'

        return ident




REMOTE_USER = 'proxyuser'
REMOTE_HOST = 'remote-host.org'
REMOTE_PORT = 22
REMOTE_KEY = '~/.ssh/id_rsa'
SSH_OPTS = ''

FORWARD = [
    {
        'local': {
            'gateway': False,            # If you want to forward anything from the gateway interface
            'host': '127.0.0.1',         # Local src address
            'port': 8015                 # Local port to forward to remote
        },
        'remote': {
            'gateway': False,     # Bind to a gateway interface (ssh host visible from internet)
            'host': '127.0.0.1',  # IP address to bind to, the tunnel will be
                                  # visible on it (`telnet 127.0.0.1 80` on remote)
            'port': 80            # Port reachable on the remote host
        },
        'validate': {
            # remote_port_ping requires "nc" to be installed on remote host
            'method': 'remote_port_ping',            # Just a check if port is open, you can place there a callback
            'interval': 60,                          # Interval for checking the health and general status
            'wait_time_before_restart': 60,          # After failure wait this time before doing restart,
                                                     # maybe the tunnel will be back
            'kill_existing_tunnel_on_failure': True  # Exit existing tunnel if it is not working
            # 'notify_url': 'http://some-slack-webhook-url'
        },
        'mode': 'remote'  # Forward remote service to be visible locally (options: local, remote)
    }
]




import os
import sys
import unittest
from typing import Tuple
from unittest.mock import Mock, patch

sys.path.append(os.path.dirname(__file__) + "/../tunman")

from ..tunman.model import HostTunnelDefinitions, Forwarding, LocalPortDefinition, RemotePortDefinition
from ..tunman.manager.ssh import TunnelManager, Validation, SIGNAL_RESTART
from ..tunman.logger import setup_dummy_logger


def create_example_portmapping(config) -> tuple:
    return LocalPortDefinition(
        gateway=False,
        host='192.168.1.5',
        port=22,
        configuration=config
    ), RemotePortDefinition(
        gateway=False,
        host='192.168.5.20',
        port=2222,
        configuration=config
    )


class ManagerTest(unittest.TestCase):
    def prepare_data(self) -> Tuple[Forwarding, HostTunnelDefinitions]:
        setup_dummy_logger()

        # prepare dependencies
        config = HostTunnelDefinitions()
        local_port, remote_port = create_example_portmapping(config)

        validate = Mock()
        validate.interval = 0
        validate.wait_time_before_restart = 1
        validate.kill_existing_tunnel_on_failure = False

        fw = Forwarding(configuration=config, local=local_port, remote=remote_port,
                        validate=validate, mode='local', retries=5, use_autossh=False, health_check_connect_timeout=0,
                        warm_up_time=0, time_before_restart_at_initialization=0, wait_time_after_all_retries_failed=0)

        return fw, config

    def test_tunnel_loop_spawns_tunnel_on_died_process(self):
        """
        Scenario: is_process_alive() returns False
        Expectation: Tunnel restart will be called

        :return:
        """

        fw, config = self.prepare_data()

        with patch.object(Validation, 'is_process_alive') as is_process_alive_mock:
            is_process_alive_mock.return_value = False

            manager = TunnelManager()
            manager.spawn_ssh_process = Mock()
            result = manager._tunnel_loop(Mock(), fw, config, '-L 127.0.0.1:3306:192.168.1.5:3306')

            assert result == SIGNAL_RESTART

    def test_tunnel_loop_spawns_tunnel_on_health_check_failed(self):
        """
        Scenario: check_tunnel_alive() fails, it means that health check failed
        Expectation: Tunnel restart will be called

        :return:
        """
        fw, config = self.prepare_data()

        with patch.object(Validation, 'is_process_alive') as is_process_alive_mock:
            is_process_alive_mock.return_value = True

            with patch.object(Validation, 'check_tunnel_alive') as check_tunnel_alive_mock:
                check_tunnel_alive_mock.return_value = False

                manager = TunnelManager()
                manager.spawn_ssh_process = Mock()
                result = manager._tunnel_loop(Mock(), fw, config, '-L 127.0.0.1:3306:192.168.1.5:3306')

                assert result == SIGNAL_RESTART

    def test_tunnel_loop_will_not_respawn_tunnel_when_its_a_false_alarm(self):
        """
        Scenario: check_tunnel_alive() fails, it means that health check failed
                  But after giving the situation a time (1 sec configured), the service brings back automatically
        Expected: Do not take any action, if the service is back online

        :return:
        """

        fw, config = self.prepare_data()
        manager = TunnelManager()

        with patch.object(manager, '_carefully_sleep') as sleep_mock:
            sleep_mock.side_effect = [True, False]  # on False the loop should exit

            with patch.object(Validation, 'is_process_alive') as is_process_alive_mock:
                is_process_alive_mock.return_value = True

                with patch.object(Validation, 'check_tunnel_alive') as check_tunnel_alive_mock:
                    check_tunnel_alive_mock.side_effect = [False, True]

                    manager.spawn_ssh_process = Mock()
                    manager._tunnel_loop(Mock(), fw, config, '-L 127.0.0.1:3306:192.168.1.5:3306')

                    assert manager.spawn_ssh_process.call_count == 0




import os
import sys
import unittest
from ipaddress import IPv4Address
from typing import List
from unittest_data_provider import data_provider

sys.path.append(os.path.dirname(__file__) + "/../tunman")

from ..tunman.model import HostTunnelDefinitions
from ..tunman.logger import setup_dummy_logger


def create_ssh_connection_string_provider():
    return [
        # Regular, key-based
        [
            {
                'remote_host': '127.0.0.1',
                'remote_port': 22,
                'remote_password': '',
                'remote_key': '/tmp/some-key_rsa',
                'remote_user': 'riotkit',
                'ssh_opts': '-E /tmp/test.log'
            },
            {
                'ssh_executable': 'ssh'
            },
            [
                'ssh', 'riotkit@127.0.0.1', '-p 22', '-i /tmp/some-key_rsa',
                '-E /tmp/test.log'
            ],
            []
        ],

        # Password based, should not contain "-i" switch
        [
            {
                'remote_host': 'riotkit.org',
                'remote_port': 2222,
                'remote_password': 'stalin-was-a-panty-pisser',
                'remote_key': '',
                'remote_user': 'riotkit',
                'ssh_opts': '-E /tmp/test.log'
            },
            {
                'ssh_executable': '/usr/bin/ssh-other-binary'
            },
            [
                '/usr/bin/ssh-other-binary', 'riotkit@riotkit.org', '-p 2222',
                '-E /tmp/test.log'
            ],
            ['-i ', '-p 22 ']
        ],

        # Without key and custom ssh options (kwargs)
        [
            {
                'remote_host': '127.0.0.1',
                'remote_port': 22,
                'remote_password': '',
                'remote_key': '/tmp/some-key_rsa',
                'remote_user': 'riotkit',
                'ssh_opts': '-E /tmp/test.log'
            },
            {
                'ssh_executable': '', 'with_key': False, 'with_custom_opts': False
            },
            [],
            ['-E /tmp/test.log', '-i /tmp/some-key_rsa', 'ssh ']
        ]
    ]


class HostTunnelDefinitionsTest(unittest.TestCase):
    def setUp(self) -> None:
        setup_dummy_logger()

    @data_provider(create_ssh_connection_string_provider)
    def test_create_ssh_connection_string(self, definition_props: dict, method_kwargs: dict,
                                          expectations_contains: List[str], expectations_not_contains: List[str]):
        definition = HostTunnelDefinitions()

        for key, value in definition_props.items():
            definition.__setattr__(key, value)

        out = definition.create_ssh_connection_string(**method_kwargs)

        for expected_string in expectations_contains:
            assert expected_string in out, 'Expected that the output command will contain %s' % expected_string

        for unexpected_string in expectations_not_contains:
            assert unexpected_string not in out, 'Expected that the output command will not contain %s' % \
                                                 unexpected_string

    @staticmethod
    def test_get_remote_gateway():
        definition = HostTunnelDefinitions()
        definition.remote_host = 'localhost'

        assert definition.get_remote_gateway() == '127.0.0.1'

    @staticmethod
    def test_get_local_gateway():
        definition = HostTunnelDefinitions()
        definition.remote_host = '1.2.3.4'
        definition.remote_port = 2222
        definition.remote_user = 'international-workers-association'
        definition.remote_key = ''
        definition.remote_password = 'world-should-be-free'
        definition.remote_passphrase = 'social-revolution'

        gw = definition.get_local_gateway()

        assert IPv4Address(gw)

    def test_parse(self):
        definition = HostTunnelDefinitions()
        definition.variables_post_processor = lambda x, y: x

        gw = definition.get_local_gateway()
        parsed = definition.parse('Local GW is {{ local_gw }}')

        assert "Local GW is %s" % gw in parsed

    def test_create_ssh_keyscan_command(self):
        definition = HostTunnelDefinitions()
        definition.remote_port = 2222
        definition.remote_host = 'iwa-ait.org'

        out = definition.create_ssh_keyscan_command()

        assert "ssh-keyscan" in out
        assert "-p 2222" in out
        assert "iwa-ait.org" in out




import re
from typing import List


class ParsedNetworkingInformation(object):
    """
    `ip route` command parser

    Provides a more stable way of extracting network information from `ip route`, it's more safe than using AWK or SED
    Because in various configurations the columns list is different, also on different systems the `ip route` gives
    different results.
    """

    _parsed: dict

    def __init__(self, ip_route_output: str):
        self._parsed = {
            'gw_interface': '',
            'gw_interface_ip': '',
            'gw_ip': '',
            'interfaces_ip': {}
        }
        self._parse(ip_route_output)

    def _parse(self, output: str):
        as_lines = output.split("\n")

        self._parse_gateway_interface(as_lines)
        self._parse_ip_of_interfaces(as_lines)
        self._parse_ip_of_gateway_interface()

    def _parse_gateway_interface(self, output: List[str]):
        for line in output:
            if "default via" in line:
                match_dev = re.search('dev ([a-z0-9]+)', line)
                match_ip = re.search('via ([0-9.]+)', line)

                self._parsed['gw_interface'] = match_dev.group(1)
                self._parsed['gw_ip'] = match_ip.group(1)
                break

    def _parse_ip_of_interfaces(self, output: List[str]):
        for line in output:
            if " dev " not in line or " src " not in line:
                continue

            match_dev = re.search('dev ([a-z0-9]+)', line)
            match_ip = re.search('src ([0-9.]+)', line)

            if not match_ip or not match_dev:
                raise Exception('Cannot parse `ip route` line: %s' % line)

            self._parsed['interfaces_ip'][match_dev.group(1)] = match_ip.group(1)

    def _parse_ip_of_gateway_interface(self):
        self._parsed['gw_interface_ip'] = self._parsed['interfaces_ip'][self._parsed['gw_interface']]

    @property
    def gateway_interface(self) -> str:
        """
        Gateway interface name ex. eth0 if all traffic goes through eth0 (default gateway)

        `default via 192.168.0.1 dev wlp2s0 proto dhcp metric 600`

        :return:
        """

        return self._parsed['gw_interface']

    @property
    def gateway_interface_ip(self) -> str:
        """
        Interface that is a default gateway ex. eth0 ip address
        Example: 192.168.0.1 is a gateway, and 192.168.0.100 is the IP address assigned on it, it will be returned

        `192.168.0.0/24 dev wlp2s0 proto kernel scope link src 192.168.0.109 metric 600`

        :return:
        """

        return self._parsed['gw_interface_ip']

    @property
    def gateway(self) -> str:
        """
        Get the gateway IP (the route where the traffic goes through ex. a WAN interface)

        Example: 192.168.0.1

        `default via 192.168.0.1 dev wlp2s0 proto dhcp metric 600`

        :return:
        """

        return self._parsed['gw_ip']

    def get_interface_ip(self, interface_name: str) -> str:
        if interface_name not in self._parsed:
            raise KeyError('%s is not a recognized interface in `ip route` output' % interface_name)

        return self._parsed['interfaces_ip'][interface_name]




import subprocess
from typing import List
from time import sleep
from threading import RLock
from traceback import format_exc
from ..model import Forwarding, HostTunnelDefinitions
from ..logger import Logger
from ..validation import Validation
from ..notify import Notify
from .sysprocess import SystemProcessManager

SIGNAL_TERMINATE = 1
SIGNAL_RESTART = 2


class TunnelManager:
    """
    Spawns tunnels and supervises them
    """

    _signatures: List[str]
    _proc_manager: SystemProcessManager
    _sleep_time = 10
    is_terminating: bool

    def __init__(self):
        self.is_terminating = False
        self._signatures = []
        self._lock = RLock(timeout=60)
        self._starts_history = {}
        self._proc_manager = SystemProcessManager()

    def spawn_tunnel(self, definition: Forwarding, configuration: HostTunnelDefinitions):
        """
        Glues the parameters, restarts the loop on crash, handles application shutdown

        Threads: Per thread

        :param definition:
        :param configuration:
        :return:
        """

        try:
            signature = definition.create_ssh_forwarding_signature()
        except Exception as e:
            signature = 'not_working_signature'

            Logger.error('Cannot create a forwarding signature, maybe an SSH error? Error says %s' % str(e))
            Logger.error(format_exc())

        Logger.info('Created SSH args: %s' % definition.create_ssh_arguments())

        with self._lock:
            self._signatures.append(signature)

        retries_left = definition.retries

        while True:
            if retries_left == 0:
                retries_left = definition.retries
                self._carefully_sleep(definition.wait_time_after_all_retries_failed)

            try:
                signal = self.spawn_ssh_process(definition, configuration, signature)
            except:
                Logger.error(format_exc())
                self._carefully_sleep(5)
                continue

            if signal == SIGNAL_TERMINATE:
                return

            if signal != SIGNAL_RESTART:
                raise Exception('Application error, unknown signal "%s"' % str(signal))

            # should not matter, secures from too much CPU usage
            self._carefully_sleep(2)
            retries_left -= 1

    def spawn_ssh_process(self, forwarding: Forwarding,
                          configuration: HostTunnelDefinitions, signature: str) -> int:
        """
        Spawns a SSH process and delegates supervising
        After fresh run of SSH tunnel it performs initial check

        Threads: Per thread

        :param forwarding:
        :param configuration:
        :param signature:
        :return:
        """

        # remove old, died processes from the internal registry
        with self._lock:
            self._proc_manager.clean_up_already_exited_processes()

        if self.is_terminating:
            return SIGNAL_TERMINATE

        cmd = configuration.create_complete_command_with_supervision(forwarding)

        # maintain the registry
        with self._lock:
            proc = self._proc_manager.spawn(cmd)

            forwarding.on_tunnel_started()
            Notify.notify_tunnel_restarted(forwarding)

        self._carefully_sleep(forwarding.warm_up_time)

        # make a delayed retry on start
        if not Validation.is_process_alive(signature):
            stdout, stderr = self._proc_manager.communicate(proc)
            Logger.error('Cannot spawn %s, stdout=%s, stderr=%s' % (cmd, stdout, stderr))

            if not self._recover_from_error(stdout + stderr, configuration):
                self._carefully_sleep(forwarding.time_before_restart_at_initialization)

            return SIGNAL_RESTART

        Logger.info('Process for "%s" survived initialization, got pid=%i' % (signature, proc.pid))

        return self._tunnel_loop(proc, forwarding, configuration, signature)

    def _tunnel_loop(self, proc: subprocess.Popen, definition: Forwarding, configuration: HostTunnelDefinitions,
                     signature: str) -> int:
        """
        One tunnel = one thread of health monitoring and reacting

        Threads: Per thread

        :param definition:
        :param configuration:
        :param signature:
        :return:
        """

        Logger.debug('Starting monitoring loop for "%s"' % signature)

        while True:
            if not self._carefully_sleep(definition.validate.interval):
                return SIGNAL_TERMINATE

            if not self._proc_manager.wait(proc):
                Logger.error('The process just exited')
                return SIGNAL_RESTART

            Logger.debug('Running checks for signature "%s"' % signature)

            if not Validation.is_process_alive(signature):
                Logger.error('The tunnel process exited for signature "%s"' % signature)
                return SIGNAL_RESTART

            if not Validation.check_tunnel_alive(definition, configuration):
                Logger.error('The health check "%s" failed for signature "%s"' % (
                    definition.validate.method, signature))

                time_to_wait_on_health_check_failure = definition.validate.wait_time_before_restart
                sleep(time_to_wait_on_health_check_failure)

                # check if after given additional short wait time the health is OK
                if time_to_wait_on_health_check_failure and Validation.check_tunnel_alive(definition, configuration):
                    Logger.info('Tunnel "%s" was recovered with restart' % signature)
                    continue

                if definition.validate.kill_existing_tunnel_on_failure:
                    self._proc_manager.kill_process_by_signature(signature)

                return SIGNAL_RESTART

    def get_stats(self, definitions: List[Forwarding]) -> dict:
        definitions_status = {}

        for definition in definitions:
            proc = self._proc_manager.find_process_by_signature(definition.create_ssh_forwarding_signature())

            definitions_status[definition] = {
                'pid': proc.pid if proc else '',
                'is_alive': proc is not None,
                'starts_history': definition.starts_history,
                'restarts_count': definition.current_restart_count,
                'ident': definition.ident
            }

        return {
            'signatures': self._signatures,
            'status': definitions_status,
            'procs_count': self._proc_manager.get_procs_count(),
            'is_terminating': self.is_terminating
        }

    @staticmethod
    def _recover_from_error(error_message: str, config: HostTunnelDefinitions) -> bool:
        """
        :param error_message:
        :param config:
        :return: Returns True when recovery was performed
        """

        if "remote port forwarding failed for listen port" in error_message and config.restart_all_on_forward_failure:
            Logger.warning('Killing all remote SSH sessions to free up the busy port')

            config.ssh_kill_all_sessions_on_remote()
            sleep(2)

            return True

        return False

    def _carefully_sleep(self, sleep_time: int):
        for i in range(0, sleep_time):
            if self.is_terminating:
                Logger.debug('Careful sleep: got termination signal')
                return False

            sleep(1)

        return True

    def close_all_tunnels(self):
        """
        Kill all processes spawned by the TunnelManager

        Threads: Called from main thread
        :return:
        """

        self.is_terminating = True
        self._proc_manager.close_all_tunnels(self._signatures)




import sys
import logging
from copy import copy


this = sys.modules[__name__]
MAPPING = {
    'DEBUG': 37,
    'INFO': 36,
    'WARNING': 33,
    'ERROR': 31,
    'CRITICAL': 41
}

PARAM_MAPPING = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

PREFIX = '\033['
SUFFIX = '\033[0m'


class ColoredFormatter(logging.Formatter):
    def __init__(self, pattern):
        logging.Formatter.__init__(self, pattern)

    def format(self, record):
        colored_record = copy(record)
        level_name = colored_record.levelname

        seq = MAPPING.get(level_name, 37)
        colored_level_name = '{0}{1}m{2}{3}' \
            .format(PREFIX, seq, level_name, SUFFIX)
        colored_record.levelname = colored_level_name

        return logging.Formatter.format(self, colored_record)


def setup_dummy_logger():
    logger = logging.getLogger('tunman')
    for handler in logger.handlers:
        logger.removeHandler(handler)
    this.logger = logger


def setup_logger(path: str, level: str):
    """ Creates a logger instance with proper handlers configured """

    logger = logging.getLogger('tunman')
    logger.setLevel(logging.INFO)
    formatter = ColoredFormatter("[%(asctime)s][%(name)s][%(levelname)s]: %(message)s")

    logging_handler = logging.StreamHandler(sys.stdout)
    logging_handler.setFormatter(formatter)

    log_file_handler = logging.FileHandler(path, 'a+')
    log_file_handler.setFormatter(formatter)

    level = PARAM_MAPPING[level] if level in PARAM_MAPPING else PARAM_MAPPING['info']
    logger.setLevel(level)
    logging_handler.setLevel(level)
    log_file_handler.setLevel(level)

    for handler in logger.handlers:
        logger.removeHandler(handler)

    logger.addHandler(logging_handler)

    this.logger = logger


class Logger:
    @staticmethod
    def debug(msg: str):
        this.logger.debug(msg)

    @staticmethod
    def info(msg: str):
        this.logger.info(msg)

    @staticmethod
    def warning(msg: str):
        this.logger.warning(msg)

    @staticmethod
    def error(msg: str):
        this.logger.error(msg)




import paramiko
import socket
import time
from typing import Union
from traceback import format_exc
from .logger import Logger
from .network.ipparser import ParsedNetworkingInformation


class SSHClient:
    """
    Wrapper to a SSH client, adds timeouts, retries, error handling and the list of common commands
    """

    _ssh: paramiko.SSHClient
    _connection_setup: dict
    _timeout: int
    _ip_route: Union[ParsedNetworkingInformation, None]

    def __init__(self, host: str, port: int, user: str, key: str, password: str, passphrase: str, timeout: int = 15):
        self._ip_route = None
        self._timeout = timeout
        self._connection_setup = {
            'hostname': host, 'port': port, 'username': user,
            'key_filename': key, 'password': password,
            'passphrase': passphrase, 'look_for_keys': False
        }

        self._connect()

    def _connect(self):
        Logger.info('SSH internal connection is starting')
        self._ssh = paramiko.SSHClient()
        self._ssh.load_system_host_keys()
        self._ssh.connect(**self._connection_setup)

    def raw_exec_command(self, command: str, env: dict = None, retries: int = 3) -> tuple:
        try:
            stdin, stdout, stderr = self._ssh.exec_command(command, environment=env, timeout=self._timeout)
        except (socket.timeout, paramiko.ssh_exception.SSHException):
            Logger.warning('SSH command failed due to timeout, retrying')
            self._connect()
            return self.raw_exec_command(command, env, retries - 1)
        except:
            Logger.warning('SSH command not possible to execute')
            Logger.warning(format_exc())

            time.sleep(1)
            self._connect()
            return self.raw_exec_command(command, env, retries - 1)

        return stdin, stdout, stderr

    def exec(self, cmd: str, env: dict = None) -> str:
        Logger.debug('SSH cmd: %s' % cmd)
        stdin, stdout, stderr = self.raw_exec_command(cmd, env=env)
        stderr_content = stderr.read().decode('utf-8')
        stdout_content = stdout.read().decode('utf-8').strip()

        if stderr_content:
            Logger.warning('SSH stderr: %s' % stderr_content)

        Logger.debug('SSH stdout: %s' % stdout_content)

        return stdout_content

    def kill_all_sessions(self):
        """ Kill all SSH sessions on the remote, then reconnect """

        try:
            self.exec('killall sshd || true')
        except paramiko.ssh_exception.SSHException:
            pass

        self._connect()

    def get_interface_ip(self, name: str) -> str:
        return self._get_parsed_ip_route().get_interface_ip(name)

    def get_docker_host_ip(self) -> str:
        return self._get_parsed_ip_route().gateway

    def get_route_gateway(self) -> str:
        return self._get_parsed_ip_route().gateway_interface_ip

    def _get_parsed_ip_route(self) -> ParsedNetworkingInformation:
        if self._ip_route is None:
            self._ip_route = ParsedNetworkingInformation(self.exec('ip route'))

        return self._ip_route
